@@ -3,7 +3,6 @@
 package find
 
 import (
-	runCmd "github.com/gruntwork-io/terragrunt/cli/commands/run"
 	"github.com/gruntwork-io/terragrunt/cli/flags"
 	"github.com/gruntwork-io/terragrunt/cli/flags/shared"
 	"github.com/gruntwork-io/terragrunt/internal/cli"
@@ -36,7 +35,7 @@ const (
 func NewFlags(opts *Options, prefix flags.Prefix) cli.Flags {
 	tgPrefix := prefix.Prepend(flags.TgPrefix)
 
-	return cli.Flags{
+	flags := cli.Flags{
 		flags.NewFlag(&cli.GenericFlag[string]{
 			Name:        FormatFlagName,
 			EnvVars:     tgPrefix.EnvVars(FormatFlagName),
@@ -100,8 +99,11 @@ func NewFlags(opts *Options, prefix flags.Prefix) cli.Flags {
 			Usage:       "Construct the queue as if a specific command was run.",
 			Aliases:     []string{QueueConstructAsFlagAlias},
 		}),
-		shared.NewFilterFlag(opts.TerragruntOptions),
 	}
+
+	flags = flags.Add(shared.NewFilterFlags(opts.TerragruntOptions)...)
+
+	return flags
 }
 
 func NewCommand(l log.Logger, opts *options.TerragruntOptions) *cli.Command {
@@ -109,8 +111,8 @@ func NewCommand(l log.Logger, opts *options.TerragruntOptions) *cli.Command {
 
 	// Base flags for find plus backend/feature flags
 	flags := NewFlags(cmdOpts, nil)
-	flags = append(flags, runCmd.NewBackendFlags(l, opts, nil)...)
-	flags = append(flags, runCmd.NewFeatureFlags(l, opts, nil)...)
+	flags = append(flags, shared.NewBackendFlags(opts, nil)...)
+	flags = append(flags, shared.NewFeatureFlags(opts, nil)...)
 
 	return &cli.Command{
 		Name:    CommandName,
